from typing import Any, Dict

import numpy as np
from nuplan.common.actor_state.agent import Agent
from nuplan.common.actor_state.oriented_box import OrientedBox
from nuplan.common.actor_state.scene_object import SceneObjectMetadata
from nuplan.common.actor_state.state_representation import StateSE2
from nuplan.common.actor_state.state_representation import StateVector2D
from nuplan.common.actor_state.static_object import StaticObject
from nuplan.common.actor_state.tracked_objects import TrackedObjects
from nuplan.common.actor_state.tracked_objects_types import (
    AGENT_TYPES,
)
from nuplan.common.actor_state.tracked_objects_types import TrackedObjectType
from nuplan.planning.simulation.observation.observation_type import DetectionsTracks
from nuplan.planning.simulation.trajectory.trajectory_sampling import TrajectorySampling

from navsim.planning.metric_caching.metric_caching_utils import StateInterpolator
from navsim.planning.scenario_builder.navsim_scenario_utils import normalize_angle, rotate_state_se2
from navsim.planning.simulation.planner.pdm_planner.observation.pdm_observation import PDMObservation

veh_height = 1.6


def local_xyh2global_xyh(x, y, h, ego_state: StateSE2):
    global_x = x + ego_state.x
    global_y = y + ego_state.y
    global_h = normalize_angle(
        rotate_state_se2(
            StateSE2(x, y, h),
            angle=ego_state.heading,
        ).heading
    )
    return global_x, global_y, global_h


def build_lctgen_obs(lctgen_data, ego_state, key_frame_idx, timestamp) -> PDMObservation:
    length_width, traj, heading, vel = (
        lctgen_data['length_width'],
        lctgen_data['traj'],
        lctgen_data['heading'],
        lctgen_data['vel']
    )
    agent_cnt = length_width.shape[0] - 1

    state_size = 6  # (time, x, y, heading, velo_x, velo_y)
    time_horizon = 5.0  # [s]
    resolution_step = 0.5  # [s]
    interpolate_step = 0.1  # [s]
    scenario_step = 0.5  # [s]

    # sample detection tracks a 2Hz
    relative_time_s = (
            np.arange(0, (time_horizon * 1 / resolution_step) + 1, 1, dtype=float) * resolution_step
    )

    gt_indices = np.arange(
        0, int(time_horizon / scenario_step) + 1, int(resolution_step / scenario_step)
    )
    lctgen_indices = int(scenario_step / interpolate_step) * gt_indices + key_frame_idx


    gt_detection_tracks = []

    for iteration in gt_indices:
        all_agents = []
        lctgen_index = lctgen_indices[iteration]
        for agent_idx in range(1, agent_cnt + 1):
            local_x, local_y, local_h = (
                traj[lctgen_index, agent_idx, 0].item(),
                traj[lctgen_index, agent_idx, 1].item(),
                heading[lctgen_index, agent_idx, 0].item()
            )
            global_x, global_y, global_h = local_xyh2global_xyh(
                local_x, local_y, local_h, ego_state
            )
            all_agents.append(
                Agent(
                    tracked_object_type=TrackedObjectType.VEHICLE,
                    oriented_box=OrientedBox(
                        center=StateSE2(
                            x=global_x,
                            y=global_y,
                            heading=global_h
                        ),
                        length=length_width[agent_idx, 0].item(),
                        width=length_width[agent_idx, 1].item(),
                        height=veh_height
                    ),
                    velocity=StateVector2D(
                        x=vel[lctgen_index, agent_idx, 0].item(),
                        y=vel[lctgen_index, agent_idx, 1].item()
                    ),
                    metadata=SceneObjectMetadata(
                        timestamp + iteration * 500000,
                        token=f'lctgen_vehicle{agent_idx}',
                        track_id=None,
                        track_token=f'lctgen_vehicle{agent_idx}',
                    ),
                )
            )

        tracked_objects = TrackedObjects(
            all_agents
        )
        det_tracks = DetectionsTracks(
            tracked_objects=tracked_objects
        )
        gt_detection_tracks.append(det_tracks)

    detection_tracks_states: Dict[str, Any] = {}
    unique_detection_tracks: Dict[str, Any] = {}

    for time_s, detection_track in zip(relative_time_s, gt_detection_tracks):

        for tracked_object in detection_track.tracked_objects:
            # log detection track
            token = tracked_object.track_token

            # extract states for dynamic and static objects
            tracked_state = np.zeros(state_size, dtype=np.float64)
            tracked_state[:4] = (
                time_s,
                tracked_object.center.x,
                tracked_object.center.y,
                tracked_object.center.heading,
            )

            if tracked_object.tracked_object_type in AGENT_TYPES:
                # extract additional states for dynamic objects
                tracked_state[4:] = (
                    tracked_object.velocity.x,
                    tracked_object.velocity.y,
                )

            # found new object
            if token not in detection_tracks_states.keys():
                detection_tracks_states[token] = [tracked_state]
                unique_detection_tracks[token] = tracked_object

            # object already existed
            else:
                detection_tracks_states[token].append(tracked_state)

    # create time interpolators
    detection_interpolators: Dict[str, StateInterpolator] = {}
    for token, states_list in detection_tracks_states.items():
        states = np.array(states_list, dtype=np.float64)
        detection_interpolators[token] = StateInterpolator(states)

    # interpolate at 10Hz
    interpolated_time_s = (
            np.arange(0, int(time_horizon / interpolate_step) + 1, 1, dtype=float)
            * interpolate_step
    )

    interpolated_detection_tracks = []
    for time_s in interpolated_time_s:
        interpolated_tracks = []
        for token, interpolator in detection_interpolators.items():
            initial_detection_track = unique_detection_tracks[token]
            interpolated_state = interpolator.interpolate(time_s)

            if interpolator.start_time == interpolator.end_time:
                interpolated_tracks.append(initial_detection_track)

            elif interpolated_state is not None:

                tracked_type = initial_detection_track.tracked_object_type
                metadata = (
                    initial_detection_track.metadata
                )  # copied since time stamp is ignored

                oriented_box = OrientedBox(
                    StateSE2(*interpolated_state[:3]),
                    initial_detection_track.box.length,
                    initial_detection_track.box.width,
                    initial_detection_track.box.height,
                )

                if tracked_type in AGENT_TYPES:
                    velocity = StateVector2D(*interpolated_state[3:])

                    detection_track = Agent(
                        tracked_object_type=tracked_type,
                        oriented_box=oriented_box,
                        velocity=velocity,
                        metadata=initial_detection_track.metadata,  # simply copy
                    )
                else:
                    detection_track = StaticObject(
                        tracked_object_type=tracked_type,
                        oriented_box=oriented_box,
                        metadata=metadata,
                    )

                interpolated_tracks.append(detection_track)
        interpolated_detection_tracks.append(
            DetectionsTracks(TrackedObjects(interpolated_tracks))
        )

    # convert to pdm observation
    pdm_observation = PDMObservation(
        trajectory_sampling=TrajectorySampling(num_poses=50, interval_length=0.1),
        proposal_sampling=TrajectorySampling(num_poses=40, interval_length=0.1),
        map_radius=100,
        observation_sample_res=1,
    )
    pdm_observation.update_detections_tracks(interpolated_detection_tracks)
    return pdm_observation
